use std::{collections::HashMap, str::FromStr};

use rand_core::OsRng;
use tokio::sync::RwLock;
use x25519_dalek::{PublicKey, StaticSecret};

use clap::Parser;

#[derive(clap::Parser, Debug)]
struct Args {
    #[clap(
        short = 'c',
        long = "connect",
        help = "server address to connect to",
        default_value = "grpc://127.0.0.1:8877"
    )]
    endpoint: String,

    #[clap(short, long, default_value = "1")]
    workers: u8,

    #[clap(short, long, default_value = "INFO")]
    log_level: String,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let args = Args::parse();
    if args.workers < 1 {
        return Ok(());
    }

    let level = tracing::Level::from_str(&args.log_level)?;
    tracing_subscriber::FmtSubscriber::builder()
        .with_max_level(level)
        .init();

    let client = libvanity::worker::worker_client::WorkerClient::connect(args.endpoint).await?;

    let (tx, rx) = tokio::sync::mpsc::channel(128);

    let requests = std::sync::Arc::new(RwLock::new(HashMap::new()));

    let mut another = client.clone();
    let requests_writer = requests.clone();
    let pull_thread = tokio::spawn(async move {
        use libvanity::worker::request::NewOrDone::{Done, New};

        let mut tasks = another
            .poll(tokio_stream::wrappers::ReceiverStream::new(rx))
            .await
            .expect("msg")
            .into_inner();
        let requests = requests_writer;
        while let Some(msg) = tasks.message().await.unwrap() {
            if let Some(new_or_done) = msg.new_or_done {
                match new_or_done {
                    New(request) => {
                        let mut writer = requests.write().await;
                        tracing::debug!("new work received: {}", request.id);
                        writer.insert(request.id, request.pattern);
                    }
                    Done(response) => {
                        let mut writer = requests.write().await;
                        tracing::debug!("stop signal received: {}", response);
                        writer.remove(&response);
                    }
                }
            }
        }
    });

    let worker_threads = Vec::from_iter(0..args.workers).into_iter().map(|id| {
        let requests = requests.clone();
        let tx = tx.clone();
        let id = id.clone();
        tracing::trace!("spawning worker {}", id);
        tokio::spawn(async move {
            tracing::trace!("worker {} started", id);
            loop {
                let requests = requests.read().await.clone();
                if requests.is_empty() {
                    tracing::trace!("no work found on {}", id);
                    tokio::time::sleep(std::time::Duration::from_millis(10)).await;
                    continue;
                }

                let private = StaticSecret::new(&mut OsRng);
                let public = PublicKey::from(&private);

                for (id, pattern) in requests.iter() {
                    let matcher = regex::Regex::new(pattern).unwrap();
                    let matcher = matcher.clone();
                    let public_b64 = base64::encode(public.as_bytes());
                    if matcher.is_match(&public_b64) {
                        tracing::trace!("match found on {}: {}", id, public_b64);
                        let mut result = libvanity::worker::Result::default();
                        result.id = id.clone();
                        result.result = base64::encode(private.to_bytes());
                        tx.send(result).await.expect("no error");
                    }
                }
            }
        })
    });

    let _ = tokio::join!(futures::future::join_all(worker_threads), pull_thread);

    Ok(())
}
